/*
  tritoni.c
  Copyright (C) Fabio Barbon <fabio@realinfo.it>.

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

  $Log: tritoni.c,v $
  Revision 1.2  2002/12/06 17:09:36  fabio
  Sistemato sorgente. Scritto README.


*/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#define TYPE_FIFTHS 0
#define TYPE_SECONDS 1
#define TYPE_THIRDS 2

const char version[] = "$Id: tritoni.c,v 1.2 2002/12/06 17:09:36 fabio Exp $";

const char *nome_int[3] = {"quinte","seconde","terze"};
const char val_int[3] = {5,2,3};

const char *names[13] = {"C","Cd","D","Dd","E","F","Fd","G","Gd","A","Ad","B"};
const char *lylnames[13] = {"c","cis","d","dis","e","f","fis","g","gis","a","ais","b"};

const char *intnames[13] = {"1","m","M","m","M","g","t","g","a","M","m","M"};
const char notes[7] = {'c','d','e','f','g','a','b'};
const int semitones[7] = {2,2,1,2,2,2,1};

int nst[13],ncnt=0,gcnt=0,arita=7,tone=0,int_type=TYPE_FIFTHS,outmode=0,
  exact_rep=0;
int rot_arr[1000][13],rot_n=0;

/*
  st_cmp

  calcola la presenza di cicli nominali nella scala corrente (nst).
*/
int st_cmp(int n)
{
  int i;
  for(i=1;i<ncnt-1;i++)
    if((nst[i]%12)==(n%12))
      return 1;
  return 0;
}

/*
  code_note
  
  Data una "nota" (intesa come distanza in semitoni dal tono di base) e
  un numero di grado di scala, calcola l'altezza e le alterazioni.
*/
int code_note(int *alter, int nota, int grado)
{
  int tonealt,cnt,altezza,i,giusta;
  
  // calcola altezza tono di base (e' stupido rifarlo ogni volta!)
  tonealt=0;
  for(cnt=0;cnt<tone;cnt += semitones[tonealt%7])
    tonealt++;

  // calcola  altezza (do,re,mi,...)
  altezza = ((val_int[int_type]-1)*grado + tonealt) % 7;

  // calcola valore nota giusta corrispondente
  giusta=0;
  for(i=0;i<altezza;i++)
    giusta += semitones[i];
  
  // calcola alterazioni (b, bb, #, ##,...)
  *alter = (nota-giusta);
  if(*alter<-6)
    *alter += 12;
  else if(*alter>+6)
    *alter -= 12;

  /*  
  if(*alter<-3 || *alter>3)
    fprintf(stderr,"Warning: alterazione eccessiva (%d)\n",*alter);
  */

  return altezza;
}



/*
  st_print

  stampa la scala corrente. 
  Sono previsti due tipi di output: a video o come sorgente lilypond.
*/
void st_print(void)
{
  int i,tmp,j,nt;
  int curalt,curalter;
  char note[10];
  int norm_seq[13];

  // calcola sequenza normalizzata
  //  nst[ncnt-1] -= 12;
  for(i=0;i<ncnt;i++)
    norm_seq[i]=nst[i];
  for(i=0;i<ncnt;i++)
    for(j=i+1;j<ncnt;j++)
      if(norm_seq[j]%12 < norm_seq[i]%12){
	tmp = norm_seq[j];
	norm_seq[j] = norm_seq[i];
	norm_seq[i] = tmp;
      }
  tmp = norm_seq[0];
  for(i=0;i<ncnt-1;i++)
    norm_seq[i]=norm_seq[i+1]%12;
  norm_seq[ncnt-1] = tmp%12+12;

  // Conta tritoni
  nt = 0;
  for(i=1;i<ncnt;i++)
    if(nst[i]-nst[i-1] == 6)
      nt++;

  switch(outmode){
    
  case 0: // text output
    printf("Tritoni %d\n",nt);
    printf(" ");

    for(i=0;i<ncnt;i++){
      //    printf("%0.2d(%0.2d) ",nst[i],nst[i]%12);
      //    printf("%0.2d(%s) ",nst[i],names[nst[i]%12]);
      // printf("%0.2d ",nst[i]-nst[i-1]);

      //      printf("%-2s ",names[(nst[i]+tone)%12]);
      if(exact_rep){
	curalt = code_note(&curalter,(nst[i]+tone)%12,i);
	printf("%c",notes[curalt]);
	for(j=0;j<abs(curalter);j++)
	  printf("%c",(curalter<0)?'b':'#');
	printf(" ");
      } else
	printf("%-2s ",names[(nst[i]+tone)%12]);
    }
    
    printf("       ");
    for(i=0;i<ncnt;i++){
      printf("%-2s ",names[(norm_seq[i]+tone)%12]);
    }
    
    printf("\n ");
    for(i=1;i<ncnt;i++){
      printf(" %s ",intnames[(nst[i]-nst[i-1])%12]);
    }
    printf("          ");
    for(i=1;i<ncnt-1;i++){
      printf(" %d ",norm_seq[i]%12-norm_seq[i-1]%12);
    }
    printf(" %d ",(norm_seq[ncnt-1]-norm_seq[ncnt-2]+144)%12);

    printf("\n-------------------------------------------------------\n");
    break;

    
  case 1: // lilypond score output
    printf("\\score { \\notes { \\time %d/4 ",ncnt);
    //    printf("\\score { \\notes { \\cadenzaOn ");

    printf("\\property Staff.instrument = \"\\kern-0.5cm\\fbox{%0.2d}\" ",gcnt+1);

    // scala reale
    for(i=0;i<ncnt;i++){
      //      printf("%0.2d(%0.2d) ",nst[i],nst[i]%12);
      //      printf("%0.2d(%s) ",nst[i],names[nst[i]%12]);

      // scelta chiave opportuna
      if((nst[i]+tone)<24)
	printf("\\clef bass ");
      else if ((nst[i]+tone)<48)
	printf("\\clef \"G\" ");
      else
	printf("\\clef \"G^8\" ");
      
      if(exact_rep){
	curalt = code_note(&curalter,(nst[i]+tone)%12,i);
	printf("%c",notes[curalt]);
	for(j=0;j<abs(curalter);j++)
	  printf("%s",(curalter<0)?"es":"is");
      } else
	printf("%s",lylnames[(nst[i]+tone)%12]);

      for(j=0;j<(nst[i]+tone)/12;j++)
	printf("\'");

      printf(" ");
    }

    // scala cromatica
    //    printf("\\relative %s ",lylnames[(norm_seq[0]+tone)%12]);
    printf("\\clef violin ");
    for(i=0;i<ncnt;i++){
      printf("%s'",lylnames[(norm_seq[i]+tone)%12]);
      for(j=0;j<(norm_seq[i]+tone)/12;j++)
	printf("\'");
      printf(" ");
    }

    // didascalia
    printf("} ");
    printf("\\header { piece = \"$<$");
    for(i=0;i<ncnt-1;i++)
      printf(" %s",intnames[(nst[i+1]-nst[i])%12]);
    printf(" $>$ (tritoni %d) - $<$",nt);

    for(i=1;i<ncnt-1;i++){
      printf(" %d",norm_seq[i]%12-norm_seq[i-1]%12);
    }
    printf(" %d $>$\" }",(norm_seq[ncnt-1]-norm_seq[ncnt-2]+144)%12);

    printf("}\n");

    break;
  }


  //  printf("\n\n");
  return ;
}

/*
  rotaz_comp

  calcola l'equivalenza (definita come simmetria rotazionale) 
  della scala corrente con quelle precedenti.
*/
int rotaz_comp(void)
{
  int i,j,offset,delta,delta2,num_match,idx,idx2;

  for(i=0;i<rot_n ;i++){
    // computa simmetria rotazionale di stack con i-esima

    for(offset=0;offset<arita ; offset++){
      num_match = 0;
      for(j=1;j<arita+1;j++){
	delta = nst[j] - nst[j-1];
	idx = (j+offset)%arita;
	idx2 = (j+offset+arita-1)%arita;
	delta2 = rot_arr[i][idx] - rot_arr[i][idx2];
	if(delta2 == delta)
	  num_match++;
      }
      //      printf("i:%d nummatch:%d\n",i,num_match);
      if(num_match==arita-1)
	return 0;
    }
  }
  
  // qui sicuramente nuova sequenza
  for(i=0;i<arita;i++){
    rot_arr[rot_n][i] = nst[i];
  }
  rot_n++;
  return 1;
}

/*
  tritone_next

  operatore di ricorsione (combinatoria stupida per ricoprire tutte
  le scale possibili fissati il numero di suoni della scala e il tipo
  di intervallo)

  tritone_next(0)   - base della ricorsione, ovvero inizio della scala
  tritone_next(n+1) - calcola scale a n+1 suoni partendo da quella corrente a n
*/
void tritone_next(int n)
{
  // follows activation stack
  nst[ncnt++] = n;

  if(n>0 && n%12==0){
    if((arita+1)==ncnt){
      if(rotaz_comp()){
	st_print();
	gcnt++;
      }
    }
  } else if (st_cmp(n)){
    // fail
    /*
    printf("fail: ");
    st_print();
    */
  } else {

    int alt;
    if(exact_rep)
      code_note(&alt,(nst[ncnt-1]+tone)%12,ncnt-1);
    if(!exact_rep || abs(alt)<=2){


    switch(int_type){
    case TYPE_SECONDS:
      tritone_next(n+1); // seconda minore
      tritone_next(n+2); // seconda maggiore
      break;

    case TYPE_THIRDS:
      tritone_next(n+3); // terza minore
      tritone_next(n+4); // terza maggiore
      break;

    case TYPE_FIFTHS:
      tritone_next(n+8); // quinta aumentata
      tritone_next(n+7); // quinta giusta
      tritone_next(n+6); // quinta diminuita
      break;
    }


    }
  }
  ncnt--;
}

void usage(void)
{
  printf("Uso: tritoni <opzioni>\n");
  printf("Calcola scale possibili a <count> suoni partendo da <tone> e procedendo\nper intervalli <type>.\n\n");
  printf("  -t<type>\tInterval type:\n\t\t1 - Seconde\n\t\t2 - Terze\n\t\t3 - Quinte (default)\n");
  printf("  -n<count> \tNote count (2-12) (defaults to 7)\n");
  printf("  -o<tone> \tTone offset from C (0-12) (defaults to 0)\n");
  printf("  -e \tRappresentazione esatta note (no sostituzioni enarmoniche).\n\tATTENZIONE: BROKEN (esclude valori con piu' di due accidenti ed ha altri bug).\n");
  printf("  -p \tText output (default)\n");
  printf("  -l \tLilypond output\n");
}

int main(int argc, char **argv)
{
  int c,calc_all=0;

  while((c=getopt(argc,argv,"t:n:o:pleh?")) != -1)
    switch(c){
    case 't': // intervals type
      int_type = atoi(optarg);
      if(int_type<0 || int_type>2){
	usage();
	exit(EXIT_FAILURE);
      }
      break;
    case 'n':
      arita = atoi(optarg);
      if(arita<2 || arita>12){
	usage();
	exit(EXIT_FAILURE);
      }
      break;
    case 'o':
      tone = atoi(optarg);
      if(tone<0 || tone>11){
	usage();
	exit(EXIT_FAILURE);
      }
      break;
    case 'p':
      outmode = 0;
      break;
    case 'l':
      outmode = 1;
      break;
    case 'e':
      exact_rep = 1;
      break;
    default:
      usage();
      exit (EXIT_SUCCESS);
    }

  gcnt=0;

  switch(outmode){

  case 0: //text output
    printf("Calcolo categorie su intervalli di %s con %d suoni\n",nome_int[int_type],arita);
    printf("Base simboli: %d\n",tone);
    printf("-------------------------------------------------------\n");
    tritone_next(0);
    printf("Found %d scales\n",gcnt);
    break;

  case 1: // lilypond score file
  
    //    for(int_type=0;int_type<3;int_type++){
    tritone_next(0);
    if(gcnt>0){
      printf("\\header { title = \"Topologia scale\" subtitle = \"Categorie scale possibili a %d note su %s\"   composer = \"%d categorie, base simboli %s\" tagline = \"generated from {\\bf tritoni.c} by Fabio Barbon {\\sf $<$fabio@realinfo.it$>$}\"}\n",arita,nome_int[int_type],gcnt,names[tone%12]);
    } else
      fprintf(stderr,"No scales found.\n");
    break;
  }
}

